@@ -1,9 +1,16 @@
-""" Cloudflare v4 API"""
+""" Cloudflare v4 API
+
+A Python interface Cloudflare's v4 API.
+
+See README.md for detailed/further reading.
+
+Copyright (c) 2016 thru 2024, Cloudflare. All rights reserved.
+"""
+
 import json
 import keyword
-from requests import RequestException as requests_RequestException, ConnectionError as requests_ConnectionError, exceptions as requests_exceptions
 
-from .network import CFnetwork
+from .network import CFnetwork, CFnetworkError
 from .logging_helper import CFlogger
 from .utils import user_agent, build_curl
 from .read_configs import read_configs, ReadConfigError
@@ -19,13 +26,30 @@ DEFAULT_GLOBAL_REQUEST_TIMEOUT = 5
 DEFAULT_MAX_REQUEST_RETRIES = 5
 
 class CloudFlare():
-    """ Cloudflare v4 API"""
+    """ A Python interface Cloudflare's v4 API.
+
+    :param email: Authentication email (if not provided by config methods).
+    :param key: Authentication key (if not provided by config methods).
+    :param token: Authentication token (if not provided by config methods).
+    :param certtoken: Authentication certtoken (if not provided by config methods).
+    :param debug: Debug is enabled by setting to True.
+    :param raw: Set to True to force raw responses so you can see paging.
+    :param use_sessions: The default is True; rarely needs changing.
+    :param profile: Profile name (default is "CloudFlare").
+    :param base_url: Rarely changed Cloudflare API URL.
+    :param global_request_timeout: Timeout value (default is 5 seconds).
+    :param max_request_retries: Number of retry times (default is 5 times).
+    :param http_headers: Additional HTTP headers (as a list).
+    :return: New instance of CloudFlare()
+
+    A Python interface Cloudflare's v4 API.
+    """
 
     class _v4base():
-        """ Cloudflare v4 API"""
+        """ :meta private: """
 
         def __init__(self, config):
-            """ Cloudflare v4 API"""
+            """ :meta private: """
 
             self.network = None
             self.config = config
@@ -265,14 +289,14 @@ class CloudFlare():
                 if identifiers[1] is None:
                     url = (self.base_url + '/'
                            + parts[0] + '/'
-                           + identifiers[0] + '/'
+                           + str(identifiers[0]) + '/'
                            + parts[1])
                 else:
                     url = (self.base_url + '/'
                            + parts[0] + '/'
-                           + identifiers[0] + '/'
+                           + str(identifiers[0]) + '/'
                            + parts[1] + '/'
-                           + identifiers[1])
+                           + str(identifiers[1]))
             else:
                 if identifiers[0] is None:
                     url = (self.base_url + '/'
@@ -280,16 +304,16 @@ class CloudFlare():
                 else:
                     url = (self.base_url + '/'
                            + parts[0] + '/'
-                           + identifiers[0])
+                           + str(identifiers[0]))
 
             if len(parts) > 2 and parts[2]:
                 url += '/' + parts[2]
                 if identifiers[2]:
-                    url += '/' + identifiers[2]
+                    url += '/' + str(identifiers[2])
                 if len(parts) > 3 and parts[3]:
                     url += '/' + parts[3]
                     if identifiers[3]:
-                        url += '/' + identifiers[3]
+                        url += '/' + str(identifiers[3])
                     if len(parts) > 4 and parts[4]:
                         url += '/' + parts[4]
 
@@ -299,22 +323,14 @@ class CloudFlare():
 
             try:
                 response = self.network(method, url, headers, params, data_str, data_json, files)
-            except requests_ConnectionError as e:
+            except CFnetworkError as e:
                 if self.logger:
-                    self.logger.debug('Call: requests connection exception! "%s"', e)
-                raise CloudFlareAPIError(0, 'connection error') from None
-            except requests_exceptions.Timeout as e:
-                if self.logger:
-                    self.logger.debug('Call: requests timeout exception! "%s"', e)
-                raise CloudFlareAPIError(0, 'connection timeout') from None
-            except requests_RequestException as e:
-                if self.logger:
-                    self.logger.debug('Call: requests exception! "%s"', e)
-                raise CloudFlareAPIError(0, e) from None
+                    self.logger.debug('Call: network error: %s', e)
+                raise CloudFlareAPIError(0, str(e)) from None
             except Exception as e:
                 if self.logger:
-                    self.logger.debug('Call: exception! "%s"', e)
-                raise
+                    self.logger.debug('Call: network exception! %s', e)
+                raise CloudFlareAPIError(0, 'network exception: %s' % (e)) from None
 
             # Create response_{type|code|data}
             try:
@@ -693,7 +709,7 @@ class CloudFlare():
             return response.text
 
     class _CFbase():
-        """ Cloudflare v4 API"""
+        """ :meta private: """
 
         def __init__(self, base, parts, content_type=None):
             """ Cloudflare v4 API"""
@@ -771,7 +787,7 @@ class CloudFlare():
                 raise CloudFlareAPIError(e=e) from None
 
     class _CFbaseUnused(_CFbase):
-        """ Cloudflare v4 API"""
+        """ :meta private: """
 
         def __init__(self, base, parts, content_type):
             """ Cloudflare v4 API"""
@@ -780,7 +796,7 @@ class CloudFlare():
             self._do = self._base.do_not_available
 
     class _CFbaseNoAuth(_CFbase):
-        """ Cloudflare v4 API"""
+        """ :meta private: """
 
         def __init__(self, base, parts, content_type):
             """ Cloudflare v4 API"""
@@ -830,7 +846,7 @@ class CloudFlare():
                 raise CloudFlareAPIError(e=e) from None
 
     class _CFbaseAuth(_CFbase):
-        """ Cloudflare v4 API"""
+        """ :meta private: """
 
         def __init__(self, base, parts, content_type):
             """ Cloudflare v4 API"""
@@ -840,7 +856,7 @@ class CloudFlare():
             self._valid = True
 
     class _CFbaseAuthUnwrapped(_CFbase):
-        """ Cloudflare v4 API"""
+        """ :meta private: """
 
         def __init__(self, base, parts, content_type):
             """ Cloudflare v4 API"""
@@ -850,7 +866,7 @@ class CloudFlare():
             self._valid = True
 
     class _CFbaseAuthCert(_CFbase):
-        """ Cloudflare v4 API"""
+        """ :meta private: """
 
         def __init__(self, base, parts, content_type):
             """ Cloudflare v4 API"""
@@ -876,41 +892,47 @@ class CloudFlare():
             v = v.replace('-','_')
         return v
 
-    def add(self, t, p1, p2=None, p3=None, p4=None, p5=None, content_type=None):
-        """ add api call to class"""
+    def add_carefully(self, t, *parts, content_type=None):
+        """ add_carefully()
+        """
+        self.add(t, parts, content_type, auto=False)
 
-        a = []
-        if p1:
-            a += p1.split('/')
-        if p2:
-            a += p2.split('/')
-        if p3:
-            a += p3.split('/')
-        if p4:
-            a += p4.split('/')
-        if p5:
-            a += p5.split('/')
+    def add(self, t, *parts, content_type=None, auto=True):
+        """ add()
 
-        parts = [p for p in [p1, p2, p3, p4, p5] if p is not None]
+        :param t: type of API call.
+        :param p1: part1 of API call.
+        :param p2: part1 of API call.
+        :param p3: part1 of API call.
+        :param p4: part1 of API call.
+        :param p5: part1 of API call.
+        :param content_type: optional value for the HTTP Content-Type for an API call.
+
+        add() is the core fuction that creates a new API endpoint that can be called later on.
+        """
+
+        api_sections = []
+        for p in parts:
+            api_sections += p.split('/')
 
         branch = self
-        element = None
-        for element in a[0:-1]:
+        for api_part in api_sections[0:-1]:
             try:
-                branch = getattr(branch, CloudFlare.sanitize_verb(element))
+                branch = getattr(branch, CloudFlare.sanitize_verb(api_part))
             except AttributeError:
-                # missing path - should never happen unless api_v4 is a busted file
-                branch = None
-                break
+                # missing path - should never happen unless api_v4 is a busted file or add_all() used
+                if not auto:
+                    raise CloudFlareAPIError(0, 'api load: api_part **%s** missing when adding path /%s' % (api_part, '/'.join(api_sections))) from None
+                # create intermediate path as required
+                f = self._CFbaseUnused(self._base, parts, content_type=None)
+                setattr(branch, CloudFlare.sanitize_verb(api_part), f)
+                branch = getattr(branch, CloudFlare.sanitize_verb(api_part))
 
-        if not branch:
-            raise CloudFlareAPIError(0, 'api load: element **%s** missing when adding path /%s' % (element, '/'.join(a)))
-
-        name = a[-1]
+        api_part = api_sections[-1]
         try:
-            branch = getattr(branch, CloudFlare.sanitize_verb(name))
+            branch = getattr(branch, CloudFlare.sanitize_verb(api_part))
             # we only are here becuase the name already exists - don't let it overwrite - should never happen unless api_v4 is a busted file
-            raise CloudFlareAPIError(0, 'api load: duplicate name found: %s/**%s**' % ('/'.join(a[0:-1]), name))
+            raise CloudFlareAPIError(0, 'api load: duplicate api_part found: %s/**%s**' % ('/'.join(api_sections[0:-1]), api_part))
         except AttributeError:
             # this is the required behavior - i.e. it's a new node to create
             pass
@@ -929,10 +951,16 @@ class CloudFlare():
             # should never happen
             raise CloudFlareAPIError(0, 'api load type mismatch')
 
-        setattr(branch, CloudFlare.sanitize_verb(name), f)
+        setattr(branch, CloudFlare.sanitize_verb(api_part), f)
 
     def find(self, cmd):
-        """ find """
+        """ find()
+
+        :param cmd: API in slash format
+        :return: fuction to call for that API
+
+        You can use this call to convert a string API command into the actual function call
+        """
         m = self
         for verb in cmd.split('/'):
             if verb == '' or verb[0] == ':':
@@ -944,11 +972,16 @@ class CloudFlare():
         return m
 
     def api_list(self):
-        """ recursive walk of the api tree returning a list of api calls"""
+        """ api_list()
+
+        :return: list of API calls
+
+        A recursive walk of the api tree returning a list of api calls
+        """
         return self._api_list(m=self)
 
     def _api_list(self, m=None, s=''):
-        """ recursive walk of the api tree returning a list of api calls"""
+        """ :meta private: """
         w = []
         for n in sorted(dir(m)):
             if n[0] == '_':
@@ -983,12 +1016,16 @@ class CloudFlare():
         return w
 
     def api_from_openapi(self, url=None):
-        """ Cloudflare v4 API"""
+        """ api_from_openapi()
+
+        :param url: OpenAPI URL or None if you use the built official URL
+
+        """
 
         return self._base.api_from_openapi(url)
 
     def __init__(self, email=None, key=None, token=None, certtoken=None, debug=False, raw=False, use_sessions=True, profile=None, base_url=None, global_request_timeout=None, max_request_retries=None, http_headers=None):
-        """ Cloudflare v4 API"""
+        """ :meta private: """
 
         self._base = None
 
@@ -1032,6 +1069,14 @@ class CloudFlare():
         if http_headers is not None:
             if not isinstance(http_headers, list):
                 raise TypeError('http_headers is not a list')
+            for h in http_headers:
+                try:
+                    t, v = h.split(':', 1)
+                except ValueError:
+                    # clearly a bad header syntax
+                    raise TypeError('http_headers bad syntax') from None
+                if len(t.strip()) == 0:
+                    raise TypeError('http_headers bad syntax') from None
             config['http_headers'] = http_headers
 
         # we do not need to handle item.call values - they pass straight thru
@@ -1051,30 +1096,30 @@ class CloudFlare():
             raise e
 
     def __del__(self):
-        """ Network for Cloudflare API"""
+        """ :meta private: """
 
         if self._base:
             del self._base
             self._base = None
 
     def __call__(self):
-        """ Cloudflare v4 API"""
+        """ :meta private: """
 
         raise TypeError('object is not callable')
 
     def __enter__(self):
-        """ Cloudflare v4 API"""
+        """ :meta private: """
         return self
 
     def __exit__(self, t, v, tb):
-        """ Cloudflare v4 API"""
+        """ :meta private: """
         if t is None:
             return True
         # pretend we didn't deal with raised error - which is true
         return False
 
     def __str__(self):
-        """ Cloudflare v4 API"""
+        """ :meta private: """
 
         if self._base.api_email is None:
             s = '["%s","%s"]' % (self._base.profile, 'REDACTED')
@@ -1083,7 +1128,7 @@ class CloudFlare():
         return s
 
     def __repr__(self):
-        """ Cloudflare v4 API"""
+        """ :meta private: """
 
         if self._base.api_email is None:
             s = '%s,%s("%s","%s","%s","%s",%s,"%s")' % (
@@ -1100,10 +1145,22 @@ class CloudFlare():
         return s
 
     def __getattr__(self, key):
-        """ __getattr__ """
+        """ :meta private: """
 
         # this code will expand later
         if key in dir(self):
             return self[key]
         # this is call to a non-existent endpoint
         raise AttributeError(key)
+
+class Cloudflare(CloudFlare):
+    """ A Python interface Cloudflare's v4 API.
+
+    Alternate upper/lowercase version.
+    """
+
+class cloudflare(CloudFlare):
+    """ A Python interface Cloudflare's v4 API.
+
+    Alternate upper/lowercase version.
+    """
